#![no_std]
#![no_main]

mod max6675;

use core::str;

use cortex_m::delay::Delay;

use display_interface::DisplayError;

use embedded_graphics::{
    mono_font::{iso_8859_1::FONT_6X10, MonoTextStyle},
    pixelcolor::BinaryColor,
    prelude::*,
    text::{Alignment, Text},
};

use embedded_hal::digital::v2::{InputPin, ToggleableOutputPin};

use fugit::RateExtU32;

use max6675::{Error, Max6675};

use panic_halt as _;

use rp_pico::{
    entry,
    hal::{
        clocks::{self, Clock},
        gpio::PinState,
        i2c::I2C,
        usb::UsbBus,
        Sio, Timer, Watchdog,
    },
    pac::{CorePeripherals, Peripherals},
    Pins,
};

use ssd1306::{mode::BufferedGraphicsMode, prelude::*, I2CDisplayInterface, Ssd1306};

use usb_device::{bus::UsbBus as UsbBusTrait, bus::UsbBusAllocator, prelude::*};

use usbd_serial::{SerialPort, USB_CLASS_CDC};

fn print_serial<'a, B: UsbBusTrait>(
    serial: &mut SerialPort<'a, B>,
    msg: &str,
) -> Result<(), UsbError> {
    let buf = msg.as_bytes();
    let mut pos = 0;

    while pos < buf.len() {
        match serial.write(&buf[pos..]) {
            Ok(count) => {
                pos += count;
                continue;
            }
            Err(UsbError::WouldBlock) => continue,
            Err(err) => return Err(err),
        }
    }

    Ok(())
}

fn println_serial<'a, B: UsbBusTrait>(
    serial: &mut SerialPort<'a, B>,
    msg: &str,
) -> Result<(), UsbError> {
    print_serial(serial, msg)?;
    print_serial(serial, "\r\n")
}

fn stringify_display_error(err: DisplayError) -> &'static str {
    match err {
        DisplayError::InvalidFormatError => "invalid format",
        DisplayError::BusWriteError => "bus write error",
        DisplayError::DCError => "dc error",
        DisplayError::CSError => "cs error",
        DisplayError::DataFormatNotImplemented => "data format not implemented",
        DisplayError::RSError => "rs error",
        DisplayError::OutOfBoundsError => "out of bounds",
        _ => "unknown error",
    }
}

fn display_text<DI, SIZE>(
    display: &mut Ssd1306<DI, SIZE, BufferedGraphicsMode<SIZE>>,
    text: &str,
) -> Result<(), DisplayError>
where
    DI: WriteOnlyDataCommand,
    SIZE: DisplaySize,
{
    let (w, h) = display.dimensions();

    Text::with_alignment(
        text,
        Point::new((w / 2).into(), (h / 2).into()),
        MonoTextStyle::new(&FONT_6X10, BinaryColor::On),
        Alignment::Center,
    )
    .draw(display)?;

    display.flush()
}

#[entry]
fn main() -> ! {
    let mut p = Peripherals::take().unwrap();
    let cp = CorePeripherals::take().unwrap();

    let mut watchdog = Watchdog::new(p.WATCHDOG);
    let clocks = clocks::init_clocks_and_plls(
        rp_pico::XOSC_CRYSTAL_FREQ,
        p.XOSC,
        p.CLOCKS,
        p.PLL_SYS,
        p.PLL_USB,
        &mut p.RESETS,
        &mut watchdog,
    )
    .ok()
    .unwrap();

    let mut delay = Delay::new(cp.SYST, clocks.system_clock.freq().to_Hz());

    let sio = Sio::new(p.SIO);
    let pins = Pins::new(p.IO_BANK0, p.PADS_BANK0, sio.gpio_bank0, &mut p.RESETS);

    let usb_bus = UsbBusAllocator::new(UsbBus::new(
        p.USBCTRL_REGS,
        p.USBCTRL_DPRAM,
        clocks.usb_clock,
        true,
        &mut p.RESETS,
    ));

    let mut serial = SerialPort::new(&usb_bus);

    let mut usb_dev = UsbDeviceBuilder::new(&usb_bus, UsbVidPid(0x2342, 0x1337))
        .manufacturer("lujoga")
        .product("Soldering Iron")
        .device_class(USB_CLASS_CDC)
        .build();

    for _ in 0..1000 {
        delay.delay_ms(5);

        if !usb_dev.poll(&mut [&mut serial]) {
            continue;
        }

        let mut buf = [0u8];
        match serial.read(&mut buf[..]) {
            Ok(_) => break,
            _ => (),
        }
    }

    let _ = println_serial(&mut serial, "Hello, world!");

    let i2c = I2C::i2c0(
        p.I2C0,
        pins.gpio0.into_mode(),
        pins.gpio1.into_mode(),
        400.kHz(),
        &mut p.RESETS,
        clocks.peripheral_clock.freq(),
    );

    let mut display = Ssd1306::new(
        I2CDisplayInterface::new_custom_address(i2c, 0x3c),
        DisplaySize128x64,
        DisplayRotation::Rotate0,
    )
    .into_buffered_graphics_mode();

    let result = display.init().map_err(stringify_display_error);
    if let Err(msg) = result {
        let _ = print_serial(&mut serial, msg);
        panic!();
    }

    let so = pins.gpio6.into_floating_input();
    let cs = pins.gpio7.into_push_pull_output_in_state(PinState::High);
    let sck = pins.gpio8.into_push_pull_output();

    let mut thermocouple = Max6675::new(cs, sck, so);
    let timer = Timer::new(p.TIMER, &mut p.RESETS);

    let mut relay = pins.gpio14.into_push_pull_output_in_state(PinState::Low);
    let button = pins.gpio15.into_pull_up_input();

    let mut led = pins.led.into_push_pull_output_in_state(PinState::Low);

    let mut counter = 0;
    loop {
        delay.delay_ms(5);
        counter = (counter + 1) % 100;

        if counter == 0 {
            led.toggle().unwrap();

            let mut bytes: [u8; 7] = [32, 32, 32, 32, 194, 176, 67];
            let text = match thermocouple.read_temperature(&mut timer.count_down()) {
                Ok(mut temperature) => {
                    temperature >>= 2;

                    for i in (0..3).rev() {
                        bytes[i] = 48u8 + (temperature % 10) as u8;
                        temperature /= 10;

                        if temperature == 0 {
                            break;
                        }
                    }

                    str::from_utf8(&bytes).unwrap()
                }
                Err(Error::IOError) => "i/o error",
                Err(Error::OpenThermocouple) => "n/c",
            };

            let _ = println_serial(&mut serial, text);

            display.clear_buffer();

            if let Err(err) = display_text(&mut display, text) {
                let _ = println_serial(&mut serial, stringify_display_error(err));
            }
        }

        if button.is_low().unwrap() {
            relay.toggle().unwrap();
            delay.delay_ms(1000);
        }

        if !usb_dev.poll(&mut [&mut serial]) {
            continue;
        }

        let mut buf = [0u8; 64];
        match serial.read(&mut buf[..]) {
            Ok(count) => {
                let _ = serial.write(&buf[..count]);
            }
            _ => (),
        }
    }
}
